// COPYRIGHT (c) 2002 PLOTNIX, INC. All rights reserved

package com.plotnix.jex;

import java.util.Iterator;
import java.util.HashMap;

import com.plotnix.jex.adapters.CompilingInterpreter;

/**
 * An interpreter directly evaluates an expression string.
 * It validates the syntax of the expression and throws an exception
 * if the syntax is invalid.
 * <p>
 * In order to introduce a new interpreter, create a class called
 * "com.plotnix.jex.<i>language-prefix</i>.Interpreter" that inherits from
 * this class.  Jex will find the class by the language prefix and 
 * invoke it to interpret expressions in that language.
 *
 * @author Dmitri Plotnikov
 * @version $Revision$
 */
public abstract class Interpreter {
    
    /**
     * Serves as a placeholder for a missing interpreter.  We don't want
     * to keep retrying to find the same interpreter.
     */
    private static final Object KNOWN_MISSING_INTERPRETER = new Object();
    
    private static HashMap interpreterImplMap = new HashMap();
    private static HashMap interpreterMap = new HashMap();
    
    /**
     * Locates and returns an Interpreter for the specified language.
     * If there is not such interpreter, looks up a Compiler
     * for this language. If one is found, returns an adapter Interpreter
     * that makes a compiler look like an Interpreter.
     * If neither an Interpreter nor a Compiler exist for the specified
     * language, returns null.
     * 
     * @param language is the name of an expression language (no colon
     *  at the end)
     * @return the Interpreter for that language
     */
    public static final Interpreter getInterpreter(String language){
        Object interpreter = interpreterMap.get(language);
        if (interpreter == null){
            interpreter = getInterpreterImplementation(language);
            if (interpreter == null){
                Compiler compiler = 
                        Compiler.getCompilerImplementation(language);
                if (compiler != null){
                    interpreter = new CompilingInterpreter(compiler);
                }
            }
            if (interpreter == null){
                interpreter = KNOWN_MISSING_INTERPRETER;
            }
            interpreterMap.put(language, interpreter);
        }        
        if (interpreter == KNOWN_MISSING_INTERPRETER){
            return null;
        }
        return (Interpreter) interpreter;
    }
    
    /**
     * Locates and returns an Interpreter for the specified language.
     * If no Interpreter exists for the specified language, returns null.
     * 
     * @param language is the name of an expression language (no colon
     *  at the end)
     * @return the Interpreter for that language
     */
    protected static final Interpreter getInterpreterImplementation(
            String language){
        Object interpreter = interpreterImplMap.get(language);
        if (interpreter == null){
            Class clazz = null;
            try {
                clazz = Class.forName(
                    "com.plotnix.jex." + language + ".Interpreter");
                interpreter = (Interpreter) clazz.newInstance();
            }
            catch (ClassNotFoundException ex){
                interpreter = KNOWN_MISSING_INTERPRETER;
            }
            catch (Exception ex){
                // TBD: wrap
                throw new RuntimeException(ex.getMessage());
            }
            interpreterImplMap.put(language, interpreter);
        }        
        if (interpreter == KNOWN_MISSING_INTERPRETER){
            return null;
        }
        return (Interpreter) interpreter;
    }
    
    /**
     * Evaluates the expression in the supplied context.
     * 
     * @param context is an evaluation context.
     * @param expression is the expression to be evaluated
     * @param arguments are argument of the expression. Can be null
     *  if the expression does not take any arguments.
     * @return the result of the evaluation of this expression
     */
    public abstract Object evaluate(Context context, String expression, 
            Object[] arguments);

    /**
     * Evaluates the expression and casts the result to the supplied type.
     * 
     * @see #evaluate(Context, String, Object[])
     */
    public abstract Object evaluate(Context context, String expression, 
            Object[] arguments, Class valueType);

    /**
     * Produces the iterator described by the expression in the
     * specified context.
     * 
     * @param context is an evaluation context.
     * @param expression is the expression to be evaluated
     * @param arguments are argument of the expression. Can be null
     *  if the expression does not take any arguments.
     * @return the result of the evaluation of this expression
     */
    public abstract Iterator iterate(Context context, String expression, 
            Object[] arguments);
    
    /**
     * Modifies a value in the specified context according to the
     * expression and the supplied arguments.  Casts the type of 
     * the value if necessary.
     * 
     * @param context is an evaluation context.
     * @param expression is the expression to be evaluated
     * @param value is the new value
     * @param arguments are argument of the expression. Can be null
     *  if the expression does not take any arguments.
     */
    public abstract void assign(Context context, String expression, 
            Object value,  Object[] arguments);

    /**
     * Checks the syntax of an expression intended for evaluation
     * with {@link #evaluate evaluate} and throws an exception
     * if there is an error.
     * 
     * @param expression is the expression to be checked
     */
    public void validateExpression(String expression){
    }

    /**
     * Checks the syntax of an expression intended for evaluation
     * with {@link #iterate iterate} and throws an exception
     * if there is an error.
     * 
     * @param expression is the expression to be checked
     */
    public void validateIterationExpression(String expression){
    }
    
    /**
     * Checks the syntax of an expression intended for evaluation
     * with {@link #assign assign} and throws an exception
     * if there is an error.
     * 
     * @param expression is the expression to be checked
     */
    public void validateVariableExpression(String expression){
    }
}
