// COPYRIGHT (c) 2002 PLOTNIX, INC. All rights reserved

package com.plotnix.jex;

import java.util.Iterator;
import java.util.HashMap;

/**
 * A MultilingualInterpreter determines the language for each expression and
 * then interprets it using that language's interpreter.
 * <p>
 * If the expression string passed to this interpreter starts with 
 * <code>"<i>language</i>:"</code>, the compiler strips the prefix off and
 * passes the remainder of the expression to the interpreter for 
 * the corresponding language. 
 * <p>The prefix is required if the MultilingualInterpreter's default
 * language is null.
 * <p>If the default language is not null and the expression is missing
 * a prefix, or the prefix does not match an existing compiler,
 * the multilingual interpreter passes the expression unchanged to the default 
 * language interpreter.
 * 
 * @author Dmitri Plotnikov
 * @version $Revision$ $Date$
 */
public class MultilingualInterpreter extends Interpreter {

    private String defaultLanguage;
    private Interpreter defaultInterpreter;
    private static HashMap genericInterpreters = new HashMap();

    /**
     * Returns a MultilingualInterpreter with the specified default language.
     * If the specified default language is not null, an interpreter 
     * (or a compiler) for that language should be available.
     * 
     * @param defaultLanguage is the default language or null
     * @return A MultilingualInterpreter
     */
    public static final Interpreter getMultilingualInterpreter(
            String defaultLanguage){
        Interpreter interpreter = (Interpreter) genericInterpreters.
                get(defaultLanguage);
        if (interpreter == null){
            interpreter = new MultilingualInterpreter(defaultLanguage);
            genericInterpreters.put(defaultLanguage, interpreter);
        }
        return interpreter;
    }
    
    private MultilingualInterpreter(String defaultLanguage){
        this.defaultLanguage = defaultLanguage;
        if (defaultLanguage != null){
            defaultInterpreter = Interpreter.getInterpreter(defaultLanguage);
            if (defaultInterpreter == null){
                throw new RuntimeException(
                    "Interpreter not available for expression language " +
                    defaultLanguage);
            }
        }
    }
    
    /**
     * Evaluates the expression using the appropriate Interpreter.
     * If the expression has the syntax "<i>lang</i>:<i>expr</i>" and
     * an interpreter is available for the language "<i>lang</i>", 
     * it passes the expression to that interpreter. Otherwise, passes the
     * whole unparsed expression to the default language interpreter.
     * 
     * @param context is the context in which the expression should be
     *      evaluated
     * @param expression is the expression to be evaluated with 
     *      an optional "lang:" prefix
     * @param arguments is an array of arguments for the expression. Can
     *      be null if the expression does not require any arguments.
     * 
     * @return the result of the evaluation
     */
    public Object evaluate(Context context, String expression, 
            Object[] arguments) {
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Interpreter interpreter = Interpreter.getInterpreter(language);
            if (interpreter != null){
                String body = expression.substring(inx + 1);
                return interpreter.evaluate(context, body, arguments);
            }
        }
        if (defaultInterpreter != null){
            return defaultInterpreter.evaluate(context, expression, arguments);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }

    /**
     * Evaluates the expression and casts the result to the supplied type.
     * 
     * @see #evaluate(Context, String, Object[])
     */
    public Object evaluate(Context context, String expression, 
            Object[] arguments, Class valueType) {
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Interpreter interpreter = Interpreter.getInterpreter(language);
            if (interpreter != null){
                String body = expression.substring(inx + 1);
                return interpreter.
                    evaluate(context, body, arguments, valueType);
            }
        }
        if (defaultInterpreter != null){
            return defaultInterpreter.
                    evaluate(context, expression, arguments, valueType);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }

    /**
     * See {@link #evaluate(Context, String, Object[]) evaluate}
     * @see Interpreter#iterate(Context, String, Object[])
     */
    public Iterator iterate(Context context, String expression, 
            Object[] arguments) {
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Interpreter interpreter = Interpreter.getInterpreter(language);
            if (interpreter != null){
                String body = expression.substring(inx + 1);
                return interpreter.iterate(context, body, arguments);
            }
        }
        if (defaultInterpreter != null){
            return defaultInterpreter.iterate(context, expression, arguments);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }

    /**
     * See {@link #evaluate(Context, String, Object[]) evaluate}
     * @see Interpreter#assign(Context, String, Object, Object[])
     */
    public void assign(Context context, String expression, 
            Object value, Object[] arguments) {
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Interpreter interpreter = Interpreter.getInterpreter(language);
            if (interpreter != null){
                String body = expression.substring(inx + 1);
                interpreter.assign(context, body, value, arguments);
                return;
            }
        }
        if (defaultInterpreter != null){
            defaultInterpreter.assign(context, expression, value, arguments);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }
    
    /**
     * Identifies the expression language and validates the expression 
     * syntax using the appropriate Interpreter
     * 
     * See {@link #evaluate(Context, String, Object[]) evaluate}
     * @see Interpreter#validateExpression(String)
     */
    public void validateExpression(String expression){
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Interpreter interpreter = Interpreter.getInterpreter(language);
            if (interpreter != null){
                String body = expression.substring(inx + 1);
                interpreter.validateExpression(expression);
                return;
            }
        }
        if (defaultInterpreter != null){
            defaultInterpreter.validateExpression(expression);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }

    /**
     * Identifies the expression language and validates the iteration
     * expression syntax using the appropriate Interpreter
     * 
     * See {@link #evaluate(Context, String, Object[]) evaluate}
     * @see Interpreter#validateIterationExpression(String)
     */
    public void validateIterationExpression(String expression){
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Interpreter interpreter = Interpreter.getInterpreter(language);
            if (interpreter != null){
                String body = expression.substring(inx + 1);
                interpreter.validateIterationExpression(expression);
                return;
            }
        }
        if (defaultInterpreter != null){
            defaultInterpreter.validateIterationExpression(expression);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }
    
    /**
     * Identifies the expression language and validates the variable
     * expression syntax using the appropriate Interpreter
     * 
     * See {@link #evaluate(Context, String, Object[]) evaluate}
     * @see Interpreter#validateIterationExpression(String)
     */
    public void validateVariableExpression(String expression){
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Interpreter interpreter = Interpreter.getInterpreter(language);
            if (interpreter != null){
                String body = expression.substring(inx + 1);
                interpreter.validateVariableExpression(expression);
                return;
            }
        }
        if (defaultInterpreter != null){
            defaultInterpreter.validateVariableExpression(expression);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }
}
